"""
Lightweight UGRID-1.0 parser

"""

import netCDF4
import numpy as np


def _valid_x(var):
    names = ["longitude", "grid_longitude", "projection_x_coordinate"]

    units = [
        "degrees_east",
        "degree_east",
        "degree_E",
        "degrees_E",
        "degreeE",
        "degreesE",
    ]
    if getattr(var, "standard_name", None) in names:
        return True
    if getattr(var, "axis", "None").lower() == "x":
        return True
    # Units are mandatory, fail if not present.
    if var.units in units:
        return True
    return False


def _valid_y(var):
    names = ["latitude", "grid_latitude", "projection_y_coordinate"]

    units = [
        "degrees_north",
        "degree_north",
        "degree_N",
        "degrees_N",
        "degreeN",
        "degreesN",
    ]

    if getattr(var, "standard_name", None) in names:
        return True
    if getattr(var, "axis", "None").lower() == "y":
        return True
    # Units are mandatory, fail if not present.
    if var.units in units:
        return True
    return False


def _mandatory_attr(var, attribute):
    if not hasattr(var, attribute):
        raise ValueError(
            f"Could not find required attribute {attribute} in {var}."
        )
    return


def get_mesh_var(nc):
    """Returns the mesh_topology variable for `nc` (netCDF4.Dataset object)."""
    mesh_var = nc.get_variables_by_attributes(cf_role="mesh_topology")
    if not mesh_var:
        raise ValueError(
            f"Could not find mesh_topology variable in the dataset {nc}"
        )
    if len(mesh_var) > 1:
        raise ValueError(
            f"Expected 1 mesh_topology variable, found {len(mesh_var)}."
        )

    mesh_var = mesh_var[0]
    _mandatory_attr(mesh_var, attribute="node_coordinates")
    _mandatory_attr(mesh_var, attribute="topology_dimension")

    if mesh_var.topology_dimension not in (1, 2):
        raise ValueError(
            f"Expected mesh dimension to be 1 or 2, got {mesh_var.topology_dimension}."
        )
    return mesh_var


def connectivity_array(connectivity, num_ind):
    """Returns the connectivity array for its correspdonding `netCDF4.Variable`
    according to UGRID-1.0.

    """
    array = connectivity[:]
    if not issubclass(array.dtype.type, np.integer):
        array = np.int_(array)
    if array.shape[0] == num_ind:
        array = array.T

    start_index = int(getattr(connectivity, "start_index", 0))
    if start_index >= 1:
        array -= start_index

    # FIXME: This won't work for more than one flag value.
    flag_values = getattr(connectivity, "flag_values", None)
    if flag_values:
        array[array == flag_values - start_index] = flag_values
    return array


def ugrid(nc):
    """Parse UGRID conventions.

    Take a netCDF4.Dataset object or a netCDF4 file/url string
    and returns a dictionary with the grid nodes, edges, and connectivy matrix.

    """
    if isinstance(nc, netCDF4.Dataset):
        pass
    else:
        nc = netCDF4.Dataset(nc)

    mesh_var = get_mesh_var(nc)
    valid_coords = (
        "node_coordinates",
        "face_coordinates",
        "edge_coordinates",
        "boundary_coordinates",
    )
    valid_connectivity = {
        "face_node_connectivity": 3,
        "face_face_connectivity": 3,
        "boundary_node_connectivity": 2,
        "edge_node_connectivity": 2,
    }

    # Used for compatibility with pyugrid.
    rename = {
        "node_coordinates": "nodes",
        "face_node_connectivity": "faces",
        "boundary_node_connectivity": "boundaries",
        "edge_node_connectivity": "edges",
    }

    grid = {}
    for key, value in mesh_var.__dict__.items():
        if key in valid_coords:
            coord_names = mesh_var.getncattr(key).strip().split()
            for name in coord_names:
                if _valid_x(nc[name]):
                    x = nc[name][:]
                elif _valid_y(nc[name]):
                    y = nc[name][:]
                else:
                    raise ValueError(
                        f"Could not recognize axis for {nc[name]}"
                    )
            grid.update({key: {"x": x, "y": y}})
        if key in valid_connectivity.keys():
            connectivity = nc[mesh_var.getncattr(key).strip()]
            num_ind = valid_connectivity[key]
            array = connectivity_array(connectivity, num_ind)
            grid.update({key: array})

    return {rename.get(k, k): v for k, v in grid.items()}



import numpy as np

from gridgeo.ugrid import ugrid


def _make_grid(coords):
    if coords.ndim != 3:
        raise ValueError(f"Expected 3 dimension array, got {coords.ndim}.")
    M, N, L = coords.shape
    polygons = np.concatenate(
        (
            coords[0:-1, 0:-1],
            coords[0:-1, 1:],
            coords[1:, 1:],
            coords[1:, 0:-1],
        ),
        axis=L,
    )

    polygons = polygons.reshape(((M - 1) * (N - 1), 4, L))
    return [p for p in polygons if not np.isnan(p).any()]


def _filled_masked(arr):
    if hasattr(arr, "filled"):
        return arr.filled(fill_value=np.NaN)
    else:
        return arr


class CFVariable(object):
    def __init__(self, nc, **kwargs):
        """
        FIXME: this should be done when slicing a CFVariable in pocean-core.
        This class is only a temporary workaround until something better is created.

        """
        self._nc = nc
        variables = self._nc.get_variables_by_attributes(**kwargs)
        if len(variables) > 1:
            raise ValueError(
                f"Found more than 1 variable with criteria {kwargs}"
            )
        elif not variables:
            raise ValueError(
                f"Could not find any variables with criteria {kwargs}"
            )
        else:
            self._variable = variables[0]
        self._coords = self._variable.coordinates.split()

    def _filter_coords(self, variables):
        valid_coords = []
        for var in variables:
            if var.name in self._coords:
                valid_coords.append(var)
        if len(valid_coords) != 1:
            raise ValueError(f"Expected a single coord, got '{valid_coords}'.")
        return valid_coords[0]

    def axis(self, name):
        return getattr(self, "{}_axis".format(name.lower()))()

    def t_axis(self):
        tvars = list(
            set(
                self._nc.get_variables_by_attributes(
                    axis=lambda x: x and str(x).lower() == "t"
                )
                + self._nc.get_variables_by_attributes(
                    standard_name=lambda x: str(x)
                    in [
                        "time"
                    ]  # We don't want coords `forecast_reference_time`.
                )
                + self._nc.get_variables_by_attributes(
                    _CoordinateAxisType=lambda x: str(x).lower() == "time"
                )
            )
        )
        # _CoordinateAxisType: Time
        return self._filter_coords(tvars)

    def crs(self):
        crs = getattr(self._variable, "grid_mapping", None)
        if crs:
            crs = self._nc[crs]
        return crs

    def x_axis(self):
        xnames = ["longitude", "grid_longitude", "projection_x_coordinate"]
        xunits = [
            "degrees_east",
            "degree_east",
            "degree_E",
            "degrees_E",
            "degreeE",
            "degreesE",
        ]
        xvars = list(
            set(
                self._nc.get_variables_by_attributes(
                    axis=lambda x: x and str(x).lower() == "x"
                )
                + self._nc.get_variables_by_attributes(
                    standard_name=lambda x: x and str(x).lower() in xnames
                )
                + self._nc.get_variables_by_attributes(
                    units=lambda x: x and str(x).lower() in xunits
                )
            )
        )
        return self._filter_coords(xvars)

    def y_axis(self):
        ynames = ["latitude", "grid_latitude", "projection_y_coordinate"]
        yunits = [
            "degrees_north",
            "degree_north",
            "degree_N",
            "degrees_N",
            "degreeN",
            "degreesN",
        ]
        yvars = list(
            set(
                self._nc.get_variables_by_attributes(
                    axis=lambda x: x and str(x).lower() == "y"
                )
                + self._nc.get_variables_by_attributes(
                    standard_name=lambda x: x and str(x).lower() in ynames
                )
                + self._nc.get_variables_by_attributes(
                    units=lambda x: x and str(x).lower() in yunits
                )
            )
        )
        return self._filter_coords(yvars)

    def z_axis(self):
        znames = [
            "atmosphere_ln_pressure_coordinate",
            "atmosphere_sigma_coordinate",
            "atmosphere_hybrid_sigma_pressure_coordinate",
            "atmosphere_hybrid_height_coordinate",
            "atmosphere_sleve_coordinate",
            "ocean_sigma_coordinate",
            "ocean_s_coordinate",
            "ocean_s_coordinate_g1",
            "ocean_s_coordinate_g2",
            "ocean_sigma_z_coordinate",
            "ocean_double_sigma_coordinate",
        ]
        zvars = list(
            set(
                self._nc.get_variables_by_attributes(
                    axis=lambda x: x and str(x).lower() == "z"
                )
                + self._nc.get_variables_by_attributes(
                    positive=lambda x: x and str(x).lower() in ["up", "down"]
                )
                + self._nc.get_variables_by_attributes(
                    standard_name=lambda x: x and str(x).lower() in znames
                )
            )
        )
        return self._filter_coords(zvars)

    def topology(self):
        vnames = ["grid_topology", "mesh_topology"]
        topologies = self._nc.get_variables_by_attributes(
            cf_role=lambda v: v in vnames
        )

        if not topologies:
            if self.x_axis().ndim == 1 and self.y_axis().ndim == 1:
                return "unknown_1d"
            elif self.x_axis().ndim == 2 and self.y_axis().ndim == 2:
                return "unknown_2d"
            else:
                raise ValueError(
                    f"Could not identify the topology for {self._nc}."
                )

        if topologies and len(topologies) > 1:
            raise ValueError(
                f"Expected 1 topology variable, got {len(topologies)}."
            )

        mesh = topologies[0]
        dims = getattr(mesh, "topology_dimension", None)
        cf_role = getattr(mesh, "cf_role", None)

        if cf_role == "mesh_topology" and dims in (1, 2):
            return "ugrid"
        if cf_role == "grid_topology" and dims == 2:
            return "sgrid"

    def polygons(self):
        if self.topology() == "ugrid":
            grid = ugrid(self._nc)
            node_x = grid["nodes"]["x"]
            node_y = grid["nodes"]["y"]
            faces = grid["faces"]
            return [(list(zip(node_x[k], node_y[k]))) for k in faces]

        if self.topology() == "sgrid":
            x, y = self.x_axis()[:], self.y_axis()[:]
            coords = np.concatenate([x[..., None], y[..., None]], axis=2)
            return _make_grid(coords)

        if self.topology() == "unknown_1d":
            x, y = self.x_axis()[:], self.y_axis()[:]
            # Some non-compliant grids, like NYHOPS, may have missing_value/fill_value.
            x = _filled_masked(x)
            y = _filled_masked(y)
            if hasattr(y, "filled"):
                y = y.filled(fill_value=np.NaN)
            x, y = np.meshgrid(x, y)
            coords = np.stack([x, y], axis=2)
            return _make_grid(coords)

        if self.topology() == "unknown_2d":
            x, y = self.x_axis()[:], self.y_axis()[:]
            # Some non-compliant grids, like NYHOPS, may have missing_value/fill_value.
            x = _filled_masked(x)
            y = _filled_masked(y)
            coords = np.concatenate([x[..., None], y[..., None]], axis=2)
            return _make_grid(coords)

    # Replication of the `netCDF4.Variable` object via composition.
    def __getitem__(self, key):
        return self._variable.__getitem__(key)

    def __repr__(self):
        return self._variable.__repr__()

    @property
    def units(self):
        return self._variable.units

    @property
    def standard_name(self):
        return self._variable.standard_name

    @property
    def long_name(self):
        return self._variable.long_name

    @property
    def coordinates(self):
        return self._variable.coordinates

    @property
    def ndim(self):
        return self._variable.ndim

    @property
    def size(self):
        return self._variable.size

    @property
    def shape(self):
        return self._variable.shape

    @property
    def scale(self):
        return self._variable.scale

    @property
    def datatype(self):
        return self._variable.datatype

    @property
    def dimensions(self):
        return self._variable.dimensions

    @property
    def dtype(self):
        return self._variable.dtype

    @property
    def field(self):
        return self._variable.field

    @property
    def name(self):
        return self._variable.name

    @property
    def mask(self):
        return self._variable.mask

    @property
    def _FillValue(self):
        return self._variable._FillValue

    @property
    def _ChunkSizes(self):
        return self._variable._ChunkSizes

    @property
    def chartostring(self):
        return self._variable.chartostring

    def chunking(self):
        return self._variable.chunking()

    def endian(self):
        return self._variable.endian()

    def filters(self):
        return self._variable.filters()

    def group(self):
        return self._variable.group()

    def ncattrs(self):
        return self._variable.ncattrs()



from docopt import docopt

from gridgeo import GridGeo

__doc__ = """
Save a GeoJSON or ESRI Shapefile representation of an ocean model grid.

Usage:
    gridio URL VAR --output=OUTFILE

    gridio (-h | --help -o | --output)

Examples:
    gridio netcdf-file.nc sea_water_temperature --output=grid.shp
    gridio remote-opendap-url sea_water_temperature --output=grid.geojson

Arguments:
  directory               Configuration directory.

Options:
  -h --help                       Show this screen.
  -o OUTFILE --output=OUTFILE     Output file
"""


def parse_args():
    args = docopt(__doc__)
    return args


def cli(args):
    url = args.get("URL")
    standard_name = args.get("VAR")
    filename = args.get("--output")

    grid = GridGeo(url, standard_name=standard_name)
    grid.save(filename, float_precision=4)


def main():
    args = parse_args()
    cli(args)


if __name__ == "__main__":
    main()



import dask
import dask.array as da
from dask.local import get_sync

dask.set_options(get=get_sync)


class DaskNetCDF(object):
    def __init__(self, variable, array=None, chunks=None):
        self._variable = variable
        self.chunks = -1 if not chunks else chunks
        if array is None:
            self.array = da.from_array(variable, chunks=chunks)
        else:
            self.array = array

    def __repr__(self):
        lines = self._variable.__repr__().split("\n")
        new_lines = ["<class 'dask_netcdf.DaskNetCDF'>"]
        for line in lines[1:]:
            if "current shape" in line:
                left = line.split("=")[0]
                line = "= ".join([left, str(self.array.shape)])
            new_lines.append(line)

        return "\n".join(new_lines)

    @property
    def dtype(self):
        return self.array.dtype

    @property
    def shape(self):
        return self.array.shape

    def transpose(self, *axes):
        new_array = self.array.transpose(*axes)
        return DaskNetCDF(
            variable=self._variable, array=new_array, chunks=self.chunks
        )

    @property
    def T(self):
        return self.transpose()

    def __getitem__(self, index):
        return self.array.__getitem__(index).compute()

    def __add__(self, other):
        new_array = self.array.__add__(other)
        return DaskNetCDF(
            variable=self._variable, array=new_array, chunks=self.chunks
        )

    def __sub__(self, other):
        new_array = self.array.__sub__(other)
        return DaskNetCDF(
            variable=self._variable, array=new_array, chunks=self.chunks
        )

    def astype(self, dtype, **kwargs):
        new_array = self.array.astype(dtype, **kwargs)
        return DaskNetCDF(
            variable=self._variable, array=new_array, chunks=self.chunks
        )



from pathlib import Path

from setuptools import find_packages, setup

import versioneer

rootpath = Path(__file__).parent.absolute()


def read(*parts):
    return open(rootpath.joinpath(*parts), "r").read()


with open("requirements.txt") as f:
    require = f.readlines()
install_requires = [r.strip() for r in require]

setup(
    name="gridgeo",
    python_requires='>=3.6',
    version=versioneer.get_version(),
    description="Convert UGRID, SGRID, and non-compliant ocean model grids to geo-like objects",  # noqa
    license="BSD-3-Clause",
    long_description=f'{read("README.md")}',
    long_description_content_type="text/markdown",
    author="Filipe Fernandes",
    author_email="ocefpaf@gmail.com",
    url="https://github.com/pyoceans/gridgeo",
    keywords=["geojson", "shapefile", "ocean models", "ugrid", "sgrid"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: BSD License",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
    ],
    platforms="any",
    packages=find_packages(),
    extras_require={"testing": ["pytest"]},
    install_requires=install_requires,
    cmdclass=versioneer.get_cmdclass(),
    entry_points={"console_scripts": ["gridio = gridgeo.gridio:main"]},
)



import os
from copy import copy
from itertools import zip_longest

import netCDF4
from shapely.geometry import MultiPolygon
from shapely.ops import unary_union

from gridgeo.cfvariable import CFVariable
from gridgeo.ugrid import ugrid

try:
    from matplotlib import tri
except ImportError:
    tri = False


def set_precision(coords, precision):
    result = []
    try:
        return round(coords, int(precision))
    except TypeError:
        for coord in coords:
            result.append(set_precision(coord, precision))
    return result


class GridGeo(object):
    """
    GridGeo class takes a nc-like object (netCDF4-python or a netCDF
    file/URL) and parse the grid information.

    """

    def __init__(self, nc, **kwargs):
        """
        Return a GridGeo class.
        nc: netCDF4-python object or a netCDF file/URL string

        """

        if isinstance(nc, netCDF4.Dataset):
            pass
        else:
            nc = netCDF4.Dataset(nc)

        var = CFVariable(nc, **kwargs)

        self.x = var.x_axis()[:]
        self.y = var.y_axis()[:]
        self.mesh = var.topology()
        self.polygons = var.polygons()
        self._geo_interface = None
        self._outline = None
        self._geometry = None

        if self.mesh == "ugrid" and tri:
            grid = ugrid(nc)
            node_x = grid["nodes"]["x"]
            node_y = grid["nodes"]["y"]
            faces = grid["faces"]
            self.triang = tri.Triangulation(node_x, node_y, triangles=faces)

    @property
    def geometry(self):
        if self._geometry is None:
            self._geometry = MultiPolygon(list(zip_longest(self.polygons, [])))
        return self._geometry

    def __str__(self):
        return f"{self.mesh}"

    def __repr__(self):
        return f"<GridGeo: {self.mesh}>"

    @property
    def outline(self):
        if self._outline is None:
            self._outline = unary_union(self.geometry)
        return self._outline

    @property
    def __geo_interface__(self):
        if self._geo_interface is None:
            self._geo_interface = self.geometry.__geo_interface__
        return self._geo_interface

    def to_geojson(self, **kw):
        """
        Return a GeoJSON representation of an grid object.
        The `kw` are based on the simplestyle-spec:
        https://github.com/mapbox/simplestyle-spec/tree/master/1.1.0

        """
        title = kw.pop("title", self.mesh)
        description = kw.pop("description", "")
        marker_size = kw.pop("marker-size", "medium")
        marker_symbol = kw.pop("marker-symbol", "")
        marker_color = kw.pop("marker-color", "7e7e7e")
        stroke = kw.pop("stroke", "555555")
        stroke_opacity = kw.pop("stroke-opacity", 1)
        stroke_width = kw.pop("stroke-width", 2)
        fill = kw.pop("fill", "555555")
        fill_opacity = kw.pop("fill-opacity", 0.6)
        float_precision = kw.pop("float_precision", 6)
        geometry = copy(self.geometry.__geo_interface__)
        geometry["coordinates"] = set_precision(
            geometry["coordinates"], float_precision
        )

        geojson = {
            "type": "Feature",
            "properties": {
                "title": title,
                "description": description,
                "marker-size": marker_size,
                "marker-symbol": marker_symbol,
                "marker-color": marker_color,
                "stroke": stroke,
                "stroke-opacity": stroke_opacity,
                "stroke-width": stroke_width,
                "fill": fill,
                "fill-opacity": fill_opacity,
            },
            "geometry": geometry,
        }
        return geojson

    def save(self, filename, fmt=None, **kw):
        formats = ["shp", "geojson"]
        extension = os.path.splitext(filename)[1]

        if not fmt:
            fmt = extension.lstrip(".")

        if fmt not in formats:
            raise ValueError(f"Expected shp or geojson, got {fmt}")

        if extension.lstrip(".") != fmt:
            filename = ".".join([filename, fmt])

        if fmt == "geojson":
            import json

            geojson = self.to_geojson(**kw)
            kw = {"sort_keys": True, "indent": 4, "separators": (",", ": ")}
            with open(filename, "w") as f:
                json.dump(geojson, f, **kw)

        if fmt == "shp":
            import fiona

            name = kw.pop("name", self.mesh)

            schema = {
                "geometry": "MultiPolygon",
                "properties": {"name": f"str:{len(name)}"},
            }

            with fiona.open(filename, "w", "ESRI Shapefile", schema) as f:
                f.write(
                    {
                        "geometry": self.__geo_interface__,
                        "properties": {"name": name},
                    }
                )
